"""Front-end for all SQL-related operations."""
from contextlib import contextmanager
import itertools
import json
import typing as T
import warnings

try:  # dataframe
    import pandas as pd
except ModuleNotFoundError:
    ...
try:  # extra
    from websocket import create_connection  # pylint: disable=import-error

    WEBSOCKET_IMPORTED = True
except ModuleNotFoundError:
    WEBSOCKET_IMPORTED = False

from pycarol import Carol, PwdAuth, PwdKeyAuth
from . import bigquery


@contextmanager
def SQLSocket(carol: Carol, *args, **kw):
    """Create Socket connection to Carol.

    Args:
        carol (Carol): Carol instance.
    """
    if not WEBSOCKET_IMPORTED:
        warnings.warn(
            """
            websocket-client is not installed.
            Please install it with:
                pip install websocket-client
        """,
            Warning,
        )
        raise ModuleNotFoundError

    if not isinstance(carol.auth, (PwdAuth, PwdKeyAuth)):
        raise TypeError("Websocket only works with `PwdAuth`, `PwdKeyAuth`")

    def _build_ws_url():
        token = carol.auth._token.access_token
        return f"wss://{carol.organization}.carol.ai/websocket/live/{token}"

    ws = create_connection(_build_ws_url())
    user_id = carol.call_api("v1/users/current")["mdmId"]
    subs = {"action": "SUBSCRIBE", "type": "SQL_RESPONSE", "subtype": user_id}
    ws.send(json.dumps(subs))
    print(json.loads(ws.recv()))
    try:
        yield ws
    finally:
        ws.close()


class SQL:
    def __init__(self, carol: Carol):
        self.carol = carol
        warnings.warn("Experimental feature. The API might change without notice.")

    def query(
        self,
        query: str,
        params: T.Optional[T.List] = None,
        method: str = "sync",
        dataframe: bool = True,
        service_account: T.Optional[T.Dict[str, str]] = None,
        **kw,
    ):
        """Execute SQL Query.

        Args:
            query: `str`
                SQL Query
            params: `list` default `None`
                Params to send to Carol
            method: `str` default `sync`
                if sync runs the query synchronously and returns the result
                if socket runs the query using web socket and returns the result
            dataframe: `bool` default `True`
                if True returns a pandas dataframe
            service_account: in case you have a service account for accessing BigQuery.
            **kw: `dict`
        Returns:

        """
        methods = ("sync", "socket", "bigquery")

        params = params or []
        payload = {"preparedStatement": query, "params": params}
        if method == "socket":
            results = _socket_query(self.carol, payload, **kw)
        elif method == "sync":
            results = _sync_query(self.carol, payload)
        elif method == "bigquery":
            results = bigquery.query(self.carol, query, service_account)
        else:
            raise ValueError(f"'method' must be either: {methods}")

        if dataframe:
            return pd.DataFrame(results)
        if not dataframe and method == "bigquery":
            return [dict(row) for row in results]

        return results


def _socket_query(carol: Carol, payload, **kwargs):
    last_result = False
    results = []
    with SQLSocket(carol, **kwargs) as ws:
        resp = carol.call_api(path="v2/sql/query", method="POST", data=payload)
        while not last_result:
            resp = ws.recv()
            resp = json.loads(resp)
            result = json.loads(resp["message"])
            last_result = result["lastResult"]
            if result["success"]:
                results.extend(result["results"])
            else:
                raise ValueError(f"{result['error']}")
    return results


def _sync_query(carol: Carol, payload):
    results = carol.call_api(path="v2/sql/querySync", method="POST", data=payload)
    if not results[0]["success"]:
        raise ValueError(f"{results[0]['error']}")
    return list(itertools.chain(*[i["results"] for i in results]))



"""Back-end for BigQuery-related code."""
import re
import typing as T

from google.cloud import bigquery
from google.cloud.bigquery.job.query import QueryJob
from google.oauth2.service_account import Credentials

from .carol import Carol
from .connectors import Connectors


def get_service_account() -> T.Dict[str, str]:
    """Get BigQuery credentials from Carol."""
    ...


def query(
    carol: Carol,
    query_: str,
    service_account: T.Optional[T.Dict[str, str]] = None,
) -> QueryJob:
    """Run query for datamodel.

    Args:
        query_: BigQuery SQL query.
        service_account: in case you have a service account for accessing BigQuery.

    Returns:
        Query result.
    """
    if service_account is None:  # must call carol to get service account
        raise NotImplementedError("You must pass a service_account. Not implemented.")

    query_ = _prepare_query(carol, query_)
    client = _generate_client(service_account)
    tenant_id = carol.tenant["mdmId"]
    dataset_id = f"labs-app-mdm-production.{tenant_id}"
    job_config = bigquery.QueryJobConfig(default_dataset=dataset_id)
    return client.query(query_, job_config=job_config)


def _prepare_query(
    carol: Carol,
    query_: str,
) -> str:
    """Render template replacing variables (if any) with values.

    {{connector_name.staging_table}} is replaced by:
        `TENANTID.stg_CONNECTORID_STAGINGNAME`
    {{datamodel_name}} is replaced by `TENANTID.dm_MODELNAME`
    Variables must follow the '{{variable}}' pattern.

    Args:
        carol: Carol object.
        query_: BigQuery SQL query.

    Return:
        Query string with template rendered.
    """
    template_vars = _get_template_vars(query_)
    if len(template_vars) == 0:
        return query_

    connectors = Connectors(carol)
    connector_names = {name for name, _ in template_vars if name is not None}
    connector_map = {
        name: connectors.get_by_name(name)["mdmId"] for name in connector_names
    }

    staging_vars = filter(lambda conn_name: conn_name[0] is not None, template_vars)
    model_vars = filter(lambda conn_name: conn_name[0] is None, template_vars)

    replace_map = {}
    for connector_name, table_name in staging_vars:
        key = f"{connector_name}.{table_name}"
        connector_id = connector_map[connector_name]
        replace_map[key] = f"`stg_{connector_id}_{table_name}`"
    for _, table_name in model_vars:
        replace_map[table_name] = f"`dm_{table_name}`"

    def _replace_func(match) -> str:
        if match.group(2) in replace_map:
            return replace_map[match.group(2)]
        raise ValueError()

    return re.sub(r"({{\s*([0-9A-z\.]+)\s*}})", _replace_func, query_)


def _generate_client(service_account: T.Dict[str, str]) -> bigquery.Client:
    """Generate client from credentials."""
    credentials = Credentials.from_service_account_info(service_account)
    return bigquery.Client(project="labs-app-mdm-production", credentials=credentials)


REGEX = re.compile(
    r"{{\s*((?P<connector_name>[0-9A-z]+)(\.))?(?P<table_name>[0-9A-z]+.)\s*}}"
)


def _get_template_vars(query_: str) -> T.Set[T.Tuple[str, str]]:
    """Get all variables in the template.

    Variables follow the '{{connector_name.table_name}}' pattern. Connector name is
    optional.

    Args:
        query_: BigQuery SQL query.

    Return:
        Set with connector name (None when there is none) and staging/model name.
    """
    return {
        (match.group("connector_name"), match.group("table_name"))
        for match in REGEX.finditer(query_)
    }
